import gzip
import json
import os
import sys
import uuid
from argparse import ArgumentParser
from io import BytesIO
from math import floor
from pathlib import Path
from subprocess import check_call, check_output
from tempfile import TemporaryFile
from typing import NamedTuple, Tuple, IO

parser = ArgumentParser()
parser.add_argument("block_device", type=Path, help="Target block device")
parser.add_argument("--kernel", "-K", type=Path, help="Kernel file")
parser.add_argument("--rootfs", "-R", type=Path, help="RootFS file")

LAYOUT_PREAMBLE = """
label: gpt
unit: sectors
first-lba: 34
sector-size: 512

"""


class Partition(NamedTuple):
    number: int
    start: int
    path: Path
    name: str
    attrs: Tuple[str, ...]
    uuid: uuid.UUID

    def fp(self) -> IO[bytes]:
        if self.path.name.endswith(".gz"):
            return gzip.open(self.path, mode="rb")
        return open(self.path, mode="rb")

    def size(self) -> int:
        with self.fp() as fp:
            fp.seek(2 << 32)
            size = fp.tell()
        return floor(size / 512)

    def offset(self) -> int:
        return self.start + self.size()

    def __str__(self):
        return (
            f"{self.number} : start={self.start}, size={self.size()}, uuid=\"{self.uuid}\", name=\"{self.name}\""
        ) + (f', attrs="{",".join(self.attrs)}"' if self.attrs else '')


ROOT = Path(__file__).parent


def main():
    args = parser.parse_args()

    # noinspection PyListCreation
    partitions = [
        Partition(
            number=1,
            start=34,
            name="bl2",
            attrs=("LegacyBIOSBootable",),
            path=ROOT / "sdcard-1-bl2.bin.gz",
            uuid=uuid.uuid4(),
        )
    ]

    partitions.append(
        Partition(
            number=2,
            start=partitions[-1].offset(),
            name="u-boot-env",
            path=ROOT / "sdcard-2-ubootenv.bin.gz",
            attrs=(),
            uuid=uuid.uuid4(),
        )
    )
    partitions.append(
        Partition(
            number=3,
            start=partitions[-1].offset(),
            name="factory",
            path=ROOT / "sdcard-3-factory.bin.gz",
            attrs=(),
            uuid=uuid.uuid4(),
        )
    )
    partitions.append(
        Partition(
            number=4,
            start=partitions[-1].offset(),
            name="fip",
            path=ROOT / "sdcard-4-fip.bin.gz",
            attrs=(),
            uuid=uuid.uuid4(),
        )
    )
    partitions.append(
        Partition(
            number=5,
            start=partitions[-1].offset(),
            name="kernel",
            path=args.kernel,
            attrs=(),
            uuid=uuid.uuid4(),
        )
    )
    partitions.append(
        Partition(
            number=6,
            start=partitions[-1].offset(),
            name="recovery",
            path=ROOT / "sdcard-6-recovery.bin.gz",
            attrs=(),
            uuid=uuid.uuid4(),
        )
    )
    partitions.append(
        Partition(
            number=7,
            start=partitions[-1].offset(),
            name="rootfs",
            path=args.rootfs,
            attrs=(),
            uuid=uuid.uuid4(),
        )
    )

    with TemporaryFile(mode="w+") as stdin:
        stdin.write(LAYOUT_PREAMBLE)
        stdin.write("\n")
        for partition in partitions:
            stdin.write(str(partition))
            stdin.write("\n")
        stdin.seek(0)
        print(stdin.read())
        stdin.seek(0)
        check_call(["sfdisk", args.block_device], stdin=stdin)

    disk_parts = json.loads(check_output(["sfdisk", "-J", args.block_device], encoding="utf-8"))

    partition_table = {
        uuid.UUID(p['uuid']): p["node"] for p in disk_parts['partitiontable']['partitions']
    }

    for partition in partitions:
        device = partition_table[partition.uuid]
        print("Writing:", partition)
        with open(device, "wb") as dest, partition.fp() as src:
            counter = 0
            sys.stderr.write(f"\rWritten: {counter} bytes")
            while data := src.read(1024 * 1024):
                dest.write(data)
                counter += len(data)
                sys.stderr.write(f"\rWritten: {counter} bytes")
                sys.stderr.flush()
            sys.stderr.write(f"\n")


if __name__ == "__main__":
    main()
